# CITROS CLI

# Table of Contents
1. [Description](#Description)
2. [Prerequisites](#Prerequisites)
3. [Installation](#Installation)
4. [CLI Commands](#CLI-Commands)  
   1. [Quick Start](#Quick-Start)
   2. [init](#init)
   3. [setup-ssh](#setup-ssh)
   4. [status](#status)
   5.  [add-remote](#add-remote)
   6.  [commit](#commit)
   7.  [pull](#pull)
   8.  [push](#push)
   9.  [login](#login)
   10. [logout](#logout)
   11. [list](#list)
   12. [run](#run)
   13. [docker-login](#docker-login)
   14. [docker-build](#docker-build)
   15. [docker-build-push](#docker-build-push)
5. [Citros Repository directory and file Structure](#citros-project-file-structure) 
   1. [notebooks](#notebooks)
   2. [parameter setups](#parameter-setups)
   3. [reports](#reports)
   4. [runs](#runs)
   5. [simulations](#simulations)
   6. [workflows](#workflows)
   7. [project.json](#projectjson)
   8. [settings.json](#settingsjson)
6. [Citros Repository Configuration](#citros-project-configuration)
    1. [Adding distributions to parameter setup](#adding-distributions-to-parameter-setup)


# Description

Welcome to Citros CLI. [Citros](http://citros.io/) serves as an innovative platform for executing ROS project simulations, automating integration, and conducting in-depth performance analysis.

The Citros CLI provides an effortless interface for ROS2 developers to initiate any number of ROS simulations for a specific project with just a single command. It offers the flexibility to set parameter value distributions, creating diverse simulation environments where each run generates different parameter values. Moreover, these operations can be executed offline without relying on any external dependencies.

Citros takes its capabilities a notch higher when the user logs in. Once logged in, users can tap into the full potential of Citros, ranging from running parallel simulations in the cloud to utilizing advanced data analysis tools for performance examination. Additionally, automatic report generation is a standout feature, aiding in effortless documentation of your work. Beyond these technical perks, logging in also paves the way for collaborative work, allowing you to engage and exchange ideas with team members.

For additional information, please refer to the Citros documentation. This will provide you with comprehensive insights and detailed instructions for effective usage of Citros in general and Citros CLI in particular, and their full suite of features.

We are dedicated to enriching your ROS project simulation experience, and this package is our contribution to that cause.


# Prerequisites

- [vscode](https://code.visualstudio.com/download)
- [Docker](https://www.docker.com/)
- [Python3](https://www.python.org/downloads/)
- [git](https://git-scm.com/)

# Installation
### option 1: without code:

        $ pip install citros 

### option 2: with code:

1. clone the repo:
    
        $ git clone git@github.com:lulav/citros_cli.git

2.  Within the cloned `citros_cli` folder, Install the package to a global bin folder:

        $ python3 -m pip install .
        $ source ~/.profile

### option 3: with code and soft links
If you are developing the citros_cli package itself, than the best practice is to create a `utils` directory under a ROS project, clone the repo into it, and install the package with soft links to dev environment. I.e. from your ROS project dir:

        $ mkdir utils && cd utils
        $ git clone git@github.com:lulav/citros_cli.git
        $ cd ..
        $ python3 -m pip install -e utils/citros_cli
    
3. Environment Variables. 
   
   `citros_cli` uses several environment variables, some of which you may change according to your needs, although for the most part, the defaults are likely to be what you want. 

| ENV | Description | used in |
| --- | --- | --- |
| `CITROS_DOMAIN` | the main domain, defaults to `citros.io` | all packages |


# CLI Commands

### Quick Start:
In essence, the Citros CLI is a collection of numerous commands, but to quickly get started with running a simulation once using the default parameter values, only two straightforward commands are required:

    $ citros init
    User is not logged in. Initialzing Citros locally.
    Intialized Citros repository.
    $ citros run

The first command, `citros init`, sets up a new `.citros` repository. If you are logged in, it will clone the `.citros` directory from a remote repository.

The second command, `citros run`, executes a simulation of the provided name a designated number of times. If you don't specify a simulation name, an interactive menu will appear, letting you select from the available simulations. If the "completions" value, representing the number of times the simulation should run, isn't specified, a single instance of the simulation will be executed.

## init
The `init` command is used to initialize a Citros repository. Depending on the user's login status, this behavior varies. For logged-out users, the project initializes locally. However, logged-in users will have the `.citros` directory cloned from the Citros remote repository. If it's a new project, an empty project will be cloned.

The initialization process involves creating a `.citros` directory within your ROS project directory and generating several files and folders therein. These files are set up to allow you to run a simulation of your project with default configurations and settings. You can tailor your Citros repository to your specific needs by manually modifying these files (see the [Project Configuration](#citros-project-configuration) section below for more details).

### parameters:
parameter|description
|--|--|
|`-dir` <folder_name> | Specifies the project's working directory. Defaults to `.`|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|
|`-project_name` <proj_name> | Optional name for the project. Defaults to the last folder in the path of *dir*|

example - initializing while logged out:

    $ citros init
    User is not logged in. Initialzing Citros locally.
    Intialized Citros repository.

example - initializing while logged in:

    $ citros init
    Setting up ssh...
    Updating Citros...
    Waiting for repo to be ready...
    Citros repo successfully cloned from remote.
    Citros successfully synched with local project.
    You may review your changes via `citros status` and commit them via `citros commit`.
    Intialized Citros repository.

Note: The init command can only be executed with effect once per project. If you attempt to initialize an existing Citros repository, you will be notified that the action is redundant, and no changes will be made. Example:

    $ citros init
    The directory /workspaces/cannon has already been initialized.
    No remotes found and user is not logged in. Working offline.

To re-initialize an existing Citros repository, you must first delete the existing .citros directory for your project.

## setup-ssh
The `setup-ssh` command sets up SSH keys for secure communication with the remote Citros repository.

**Important**: If you execute `citros init` while logged in, the `setup-ssh` command will be automatically invoked behind the scenes, rendering a direct call unnecessary. However, if you were logged out during the `citros init` command execution and subsequently wish to interact with the online Citros system (e.g., running commands like `citros push`), you will need to manually run the `setup-ssh` command to establish secure online communication.

### parameters:
parameter|description
|--|--|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|

## status
The `status` command first syncs any changes in your ROS project with your Citros repository and than retrieves the current state of your Citros repository. Essentially, it acts as a wrapper for the `git status` command specifically for your Citros repository.

This command provides a quick and concise overview of the changes made to your project, giving you insights into tracked, modified, and staged files.

### parameters:
parameter|description
|--|--|
|`-dir` <folder_name> | Specifies the project's working directory. Defaults to `.`|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|

## add-remote
The `add-remote` command associates a remote Citros repository, named `origin`, with your local repository. This remote repository is hosted on the Citros servers.

### prerequisites:
`citros setup-ssh` has already been run.

**Important**: If you execute `citros init` while logged in, the `add-remote` command will automatically run in the background, making a direct call unnecessary. However, if you initially ran `citros init` while logged out and later decide to work with the online Citros system (e.g., running commands like `citros push`), you will need to manually run the `add-remote` command.

Furthermore, to ensure secure communication with the server, the `setup-ssh` command should be executed before running add-remote.

### parameters:
parameter|description
|--|--|
|`-dir` <folder_name> | Specifies the project's working directory. Defaults to `.`|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|

## commit
The `commit` command captures all modifications to your local Citros repository in a snapshot, essentially serving as a wrapper for the `git commit` command, but tailored to your Citros repository.

By executing this command, you essentially save the current state of your project, allowing you to keep track of your progress, revert changes, and even collaborate more effectively. This forms an integral part of managing and controlling the version history of your Citros repository.

### parameters:
parameter|description
|--|--|
|`-dir` <folder_name> | Specifies the project's working directory. Defaults to `.`|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|
|`-m`, `--message` | Commit message|

### example:

    $ citros commit -m "added an awesome feature"

## pull
The `pull` command fetches from and integrates with another Citros repository or a local branch. Essentially, it acts as a wrapper for the `git pull` command within the context of your Citros repo.

### parameters:
parameter|description
|--|--|
|`-dir` <folder_name> | Specifies the project's working directory. Defaults to `.`|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|
|`-b`, `--branch` | The branch name. Defaults to `master` if not given.|

### example

    $ citros pull

## push
The `push` command transfers all committed changes in your local Citros repository to the remote repository. Essentially, it acts as a wrapper for the `git push` command within the context of your Citros repo.

By employing the `push` command, you are synchronizing your local project modifications with the remote repository. This is crucial not only for backing up your work on the server but also for enabling seamless collaboration with other team members using the Citros platform.

### parameters:
parameter|description
|--|--|
|`-dir` <folder_name> | Specifies the project's working directory. Defaults to `.`|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|
|`-b`, `--branch` | The branch name. Defaults to `master` if not given.|

### example

    $ citros push

## login
The `login` command allows you to authenticate your session with Citros. To use this command, you must already have a registered account with [Citros](citros.io), including a valid username (email) and password.

By logging in, you unlock additional features such as cloud-based simulations, data analysis tools, automated report generation, and collaboration with other Citros users. Use this command to seamlessly integrate your local workspace with the Citros platform and fully utilize its capabilities.

### parameters:
parameter|description
|--|--|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|
|`-username` | The user's username (email).|
|`-password` | The user's password|

After entering the command, if either the username or password was not given, you will be prompted for your email (the username) and password.

example:

    $ citros login
    $ email: shalev@lulav.space
    $ Password: 
    User logged in.


## logout
The logout command terminates your active session with Citros.

### parameters:
parameter|description
|--|--|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|

example:

    $ citros logout
    User logged out.

## list
The `list` command displays all available simulation names. These names are derived from the filenames in the `simulations` folder within your Citros repository. Each of these files corresponds to an available launch file in your ROS project. For instance, if your ROS project contains a launch file named `foo.launch.py`, a corresponding simulation file named `simulation_foo.json` will be generated in your simulations folder.

### parameters:
parameter|description
|--|--|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|

example:

    $ citros list
    1. simulation_cannon_analytic
    2. simulation_cannon_numeric


## run
The `run` command launches a simulation either locally on your machine, or remotely on the Citros cluster.

### prerequisites:
Ensure that the project has been built and sourced, for example:
    
    $ colcon build
    $ source install/local_setup.bash

If you'd like to run your simulation remotely, you would also need to make sure:
1. You're logged in (via `citros login`).
2. You've built and pushed a docker image of your project (using `citros docker-build-push`).
3. Your `.citros` directory is synched with the remote repository (using `citros commit` and `citros push`). 

### parameters:
parameter|description
|--|--|
|`-dir` <folder_name> | Specifies the project's working directory. Defaults to `.`|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|
|`-s`, `--simulation_name` <sim_name> | Identifies the simulation you want to run. This is the name of the JSON file (excluding the `json` suffix) in the `simulations` folder. If you don't provide a simulation name, an interactive menu will display allowing you to select from the available simulations.|
|`-b`, `--batch_id` | Batch ID. Intended for Citros internal use only - DO NOT USE.|
|`-n`, `--batch_name` | Assigns a descriptive name for this simulation run, e.g. according to its settings and/or parameter setup. You can disable this option requirement via `settings.json`. If disabled, and no name is given, the default name will be the date and time.|
|`-m`, `--batch_message` | Provides a descriptive message for this simulation run, e.g. according to its settings and/or parameter setup. This can also be disabled via `settings.json`.|
|`-i`, `--run_id` | Simulation run ID. Intended for Citros internal use only - DO NOT USE.|
|`-c`, `--completions` | Sets the number of completions (simulation runs). Defaults to 1 if not specified.|
|`-t`, `--timeout` | Sets the simulation timeout in seconds. Any simulation run exceeding this time from the start will be terminated. Defaults to 60 seconds if not provided.|
|`-r`, `--remote` | Executes the simulation remotely on the cluster. See prerequisites above for details.|

If no simulation name was provided, an interactive session will begin, and you will be prompted to select a simulation from the list of available simulations (via up, down and enter keys). 

example:

    $ citros run
    ? Please choose the simulation you wish to run 
    ❯ simulation_cannon_analytic
      simulation_cannon_numeric

## docker-login
The `docker-login` command functions as a straightforward wrapper around docker login. It is designed to authenticate you with Docker Hub, which is crucial for executing the citros `docker-build-push` command.

Once authenticated, you'll have the necessary permissions to create, update, and push Docker images of your Citros repository to Docker Hub.

**Note**: You must be logged in to Citros in order to run this command.

### parameters:
parameter|description
|--|--|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|

### example

    $ citros docker-login
    WARNING! Your password will be stored unencrypted in /home/ros/.docker/config.json.
    Configure a credential helper to remove this warning. See
    https://docs.docker.com/engine/reference/commandline/login/#credentials-store

    Login Succeeded

## docker-build
The `docker-build-push` command is used to construct a Docker image of your ROS project. This image encapsulates your project's environment, facilitating the portability and reproducibility of your simulations across different systems.

### parameters:
parameter|description
|--|--|
|`-dir` <folder_name> | Specifies the project's working directory. Defaults to `.`|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|
|`-n`, `--image_name` | The requested image name (e.g. the project name). Defaults to the last folder in the path of dir |
|`-t`, `--tag` | the requested tag name for the image. Defaults to `latest`|

### example

    $ citros docker-build
    Building Docker image...
    => building with "default" instance using docker driver
    => ...
    Done.

## docker-build-push
The `docker-build-push` command is used to construct a Docker image of your ROS project and upload it to Docker Hub. This image encapsulates your project's environment, facilitating the portability and reproducibility of your simulations across different systems. 

Two tagged images will be built and pushed: `latest` and the ROS project's latest commit hash, so that it is archived in the docker registry.

### Prerequisites:
Before you can execute the docker-build-push command, you must be authenticated with Docker Hub. This can be achieved by running the `docker-login` command. 

### parameters:
parameter|description
|--|--|
|`-dir` <folder_name> | Specifies the project's working directory. Defaults to `.`|
|`-d`, `--debug` | Sets the logging level to debug.|
|`-v`, `--verbose` | Enables verbose console output.|
|`-n`, `--image_name` | The requested image name (e.g. the project name). Defaults to the last folder in the path of dir |

### example
    $ citros docker-build-push
    Building Docker image...
    => building with "default" instance using docker driver
    => ...
    Done.
    Pushing Docker image...
    The push refers to repository [us-central1-docker.pkg.dev/citros/lulav/cannon]
    ea97705925b1: Pushed 
    0d42db2cff87: Preparing 
    3752398ae296: Layer already exists 
    ...
    latest: digest: sha256:b5d109f83c1dbbaf97d918e721889988210d9bc1a91f3ecde884fbc394bcca1c size: 5136
    The push refers to repository [us-central1-docker.pkg.dev/citros/lulav/cannon]
    ea97705925b1: Layer already exists 
    5deba67bdae6: Layer already exists 
    ... 
    7f858865b89b41f493d1197c3329c0214996a625: digest: sha256:b5d109f83c1dbbaf97d918e721889988210d9bc1a91f3ecde884fbc394bcca1c size: 5136
    Done.

---

# Citros Repository file Structure

The following folder and file structure is automatically generated for you (when you run `citros init`):

- .citros
  - notebooks
  - parameter_setups
    - `default_param_setup.json`
  - reports
  - runs *(see below)*
  - simulations
    - `simulation_foo.json`
    - `simulation_bar.json`
  - workflows
    - `default_flow.json`
  - `project.json`
  - `settings.json`
  - `user_commit`

The `runs` directory will be populated with further files and folder every time you run a simulation (via `citros run`):

- runs
  - simulation name
    - batch name
      - run id
        - bag
          - `bag_0.db3`
          - `metadata.yaml`
        - config
          - `pkg1.yaml`
          - `pkg2.yaml`
        - msgs
          - `pkg1`
            - `msg`
              - `foo.msg`
          - `pkg2`
            - `msg`
              - `bar.msg`
        - `citros.log`
        - `environment.json`
        - `info.json`
        - `metrics.csv`
        - `otlp_trace`
      - `info.json`

## notebooks

TODO

## parameter setups

The `parameter_setups` directory stores your JSON-formatted parameter setup files. When you initialize your citros repository, a `default_param_setup.json` file is automatically generated. This file consolidates all the default parameters for every node across all the packages in your ROS project, providing a consolidated and easily accessible record of these parameters.

However, this `default_param_setup.json` is overwritten during citros `init`, `run`, and `status` commands. Therefore, it is recommended to duplicate this file under a different name within the `parameter_setups` directory before making any modifications. This practice ensures your custom setups are preserved and allows you to experiment with various parameter configurations.

The structured format of the parameter setup files facilitates the quick comprehension and modification of the parameters of each node in your ROS project. This is especially beneficial when you wish to explore how different parameter values or distributions impact your ROS project's behavior. For instance, you can replace a static parameter value, like 42, with a normal distribution around 42. This enables you to observe how your simulation behaves with values near 42.

To learn how to add distributions to parameter setups, please refer to the [Adding distributions to parameter setup](#Adding-distributions-to-parameter-setup) section below.

## reports

TODO

## runs

The runs directory stores data and metadata about each run of your simulations. Its structure is as follows:

- Simulation Name: These directories are named after each of the simulations defined in the simulation files. For every simulation file that is run, a corresponding directory is created here. Each Simulation Name directory may include multiple Batch Name directories.
    - Batch Name: This directory holds a batch of simulation runs. A batch consists of multiple runs of the same simulation with different parameters.
        - Run ID: Each unique simulation run has its own directory, identified by a Run ID. Under this directory, there are several files and sub-directories:
            - `bag`: This sub-directory holds the recorded data from the simulation run. It includes:
                - bag_0.db3: This is a ROS bag file that contains all the messages that were sent during the simulation.
                - metadata.yaml: A file holding metadata information associated with the bag file.
            - `config`: This sub-directory contains YAML files (pkg1.yaml, pkg2.yaml, etc.) for each package in your ROS project, detailing the actual parameters used in the simulation. If you used any distributions in your parameter setup, the values appearing here will be those that were evaluated according to the distribution you defined.
            - `msgs`: This sub-directory contains all the ROS msg files you may have in your project, each under yet another sub-directory with a name corresponding to the package the msg file belongs to.
            - `citros.log`: A standard log file that was active during the simulation run, documenting actions and events throughout the simulation.
            - `environment.json`: A file capturing a snapshot of your environment variables and Python packages at the time of the simulation run.
            - `info.json`: A JSON file containing general metadata about the run, such as batch ID, batch name, datetime of the run, user's Git commit and branch information, and Citros' Git commit and branch information, as well as a hash of the bag file.
            - `metrics.csv`: A CSV file recording system performance metrics during the simulation run, including CPU usage, total memory, available memory, used memory, and memory usage percentage.

These files collectively provide a comprehensive record of each simulation run, the conditions under which it was run, and the results it produced. This makes it easy to reproduce and understand the results of each simulation.


## simulations

The `simulations` directory stores your JSON-formatted simulation files.

A simulation json file is an auto-generated file corresponding to each launch file in your ROS project. For instance, a launch file named `foo.launch.py` will have a corresponding `simulation_foo.json` file. This file outlines the details necessary to run the corresponding simulation, specifying parameters, resources, and launch files.

Here's a breakdown of its typical structure and content:

- `description`: This is a descriptive field for the simulation setup. You can modify it to better describe your specific simulation.
- `parameter_setup`: This field points to the parameter setup JSON file that will be used for this simulation. By default, it points to `default_param_setup.json`, but you can point it to any custom parameter setup file you created in the `parameter_setups` directory.
- `launch_file`: Specifies the ROS launch file that will be used to start the simulation. For instance, `foo.launch.py`.
- `timeout`: This is the maximum time (in seconds) the simulation is allowed to run. The default is 60 seconds. If the simulation does not conclude within this timeframe, it will be terminated.
- `GPU`: Specifies the number of GPU resources required for the simulation. The default is 0, indicating that no GPU resources are needed.
- `CPU`: Specifies the number of CPU resources required for the simulation. The default is 2.
- `MEM`: Specifies the amount of memory required for the simulation, e.g., "265MB".

You can modify these fields to suit your simulation needs, just remember to save your customized version under a different name to prevent overwriting during citros `init`, `run`, or `status` commands.

## workflows

The `workflows` directory stores your JSON-formatted workflow files.

A flow.json file (e.g., `default_flow.json` which is auto-generated during `citros init`) is a user-crafted file used to automate and manage the flow of simulations in a citros repository. This file controls when the flow is triggered, which simulations are run, the post-processing analysis using Jupyter notebooks, and the recipients of the final reports. Here is a breakdown of its structure and content:

- `trigger`: This field specifies the event that initiates the flow. It is usually tied to some form of version control event, like a Git push, but can be configured according to the user's needs.
- `simulations`: This is an array of simulations to be run, specified as pairs of simulation name and the number of times to run them. For example, ["sim1", 17] means the simulation "sim1" will be run 17 times. Multiple simulations can be listed and each will be run the specified number of times.
- `notebooks`: This is a list of Jupyter notebooks used for post-processing analysis of the simulation results. For example, ["nb1.ipynb", "nb2.ipynb"] means these two notebooks will be run once the simulations complete, with the results used as their input data.
- `recipients`: This is a list of email addresses that will receive the reports generated from the notebooks' analysis.

The flow.json file helps to streamline and automate your citros repository by tying together simulation runs, data analysis, and report distribution into a single manageable file. You can customize it to suit the specifics of your project.

## project.json

The project.json file is a key component of your Citros repository. It contains metadata about your ROS project, and is automatically generated by the citros `init`, `run` and `status` commands. Here's a description of its top-level fields:

- `citros_cli_version`: the Citros CLI version installed.
- `cover`: A placeholder for a potential image that represents the project.
- `description`: A string for providing a detailed description of the project.
- `git`: The git repository URL associated with the project.
- `image`: A name that corresponds to the docker image of the project.
- `is_active`: A boolean flag indicating whether the project is active or not.
- `launches`: An array for storing metadata about launch files associated with the project. 
**Note**: these are the global launch files, which are not associated with any specific package. Generally, they are less commonly used. For package launch files, see inside the list of *packages*.
- `license`: A string indicating the license of the project.
- `name`: The name of the project. *Note*: this is the only field that you may edit and it will not be overwritten during subsequent citros commands.
- `packages`: An array of objects that describe the ROS packages that exist within the project.
- `path`: The directory path to the project.
- `readme`: The contents of the project's README file.
- `tags`: An array of strings for tagging and categorizing the project.

In the `packages` array, each object describes a specific package within the project. These objects contain similar information to the top-level fields, with additional fields:

- `maintainer`: The maintainer of the package.
- `maintainer_email`: The email address of the maintainer.
- `nodes`: An array of objects describing each node in the package, including their parameters and entry points.
- `package_xml`: The path to the package's XML file.
- `setup_py`: The path to the package's `setup.py` file. For python ROS projects only.
- `cmake`: The path to the package's `CMakeLists.txt` file. For C++ ROS projects only.

The `nodes` array contains objects that describe the ROS nodes within a package. Each object includes the following fields:

- `entry_point`: The entry point for the node, typically the function that should be executed when the node is run.
- `name`: The name of the node.
- `parameters`: An array of objects that describe the parameters associated with the node, including their name, type, and value.
- `path`: The path to the node's Python file.

## settings.json

The settings.json file holds configuration settings for your Citros repository. Here is a breakdown of each field in this file:

- `name`: The name of the current settings profile. This can be useful if you want to maintain different sets of settings for different contexts (e.g., 'default_settings', 'debug_settings', etc.).
- `bag_type`: This setting determines the storage format for the ROS bag files generated during the project's simulation runs. The possible valid value are `sqlite` (default) and `mcap`.
- `force_message`: This is a boolean setting (in string format). If set to "True", it enforces that a descriptive message is provided for each batch of simulation runs. This can be helpful for keeping track of the purpose or characteristics of each run batch.
- `force_batch_name`: Similar to force_message, this is a boolean setting (in string format). If set to "True", it enforces that a unique name is provided for each batch of simulation runs. This can be useful for organizing and identifying different batches of runs.

# Citros Repository Configuration

## Adding distributions to parameter setup
In order to define a distribution in your parameter setup file, simply replace any constant parameter value with a `distribution object`, which is demontrated in the following example: 

    {
        "distribution_type": "POISSON",
        "distribution_param1": 1.0,
        "param1_type": "FLOAT",
        "distribution_param2": 1,
        "param2_type": "INT"
    }

where `distribution_type` is one of the following:

    ["NORMAL", "EXPONITIONAL", "LAPLACE", "POISSON", "POWER", "UNIFORM", "ZIPF", "VONMISES", "RAYLEIGH", "FLOAT", "STRING"]

and `param1_type` and `param2_type` are either `"INT"` or `"FLOAT"`.

If your distribution only requires one parameter, the second will be ignored, but all 5 key/value pairs are required.

Here is an example of how you could modify your param_setup.json file:

    {
    "packages": [
        {
            "JokeTeller_Bot": [
                {
                    "HumorEngine": {
                        "ros__parameters": {
                            "joke_intensity": 11,
                            "joke_distribution": {
                                "distribution_type": "NORMAL",
                                "distribution_param1": 1.0,
                                "param1_type": "FLOAT",
                                "distribution_param2": 1,
                                "param2_type": "INT"
                            },
                            "funny_numbers": [
                                42,
                                69,
                                {
                                    "distribution_type": "POISSON",
                                    "distribution_param1": 1.0,
                                    "param1_type": "FLOAT",
                                    "distribution_param2": 1,
                                    "param2_type": "INT"
                                },
                                17
                            ]
                        }
                    }
                }
            ]
        }
    ]}
